#undef DEBUG			/* not debugging */
/****************************************************************
 *	Usage							*
 * 		t-io [options] [device1] [device2]		*
 *								*
 *	Required options (select 1 only)			*
 *		-i		Copy tape to disk 		*
 *				(device1 is the input tape)	*
 *								*
 *		-o		Copy tape to disk 		*
 *				(device1 is the output tape)	*
 *								*
 *		-c		Copy tape to tape		*
 *				(device1 is the input tape)	*
 *				(device2 is the output tape)	*
 *								*
 *		-A		Copy tape to tape (archive fmt)	*
 *				(Add a first file containing	*
 *				the blocking information)	*
 *				(device1 is the input tape)	*
 *				(device2 is the output tape)	*
 *								*
 *		-U		Unarchive (copy archive to 	*
 *					normal format)		*
 *				(device1 is the input tape)	*
 *				(device2 is the output tape)	*
 *								*
 *	Options							*
 *		-v 		Print information about 	*
 *				what is being written		*
 *								*
 *		-V#		Print information about each 	*
 *				record that is written 		*
 *				# number of records per output	*
 *								*
 *		-a		Don't abort on a read error	*
 *				(ask the user what to do)	*
 *								*
 *		-t#		Set tape number (default=0)	*
 *				Several tapes may use the	*
 *				same directory			*
 *								*
 *		-b#		use # as max blocking		*
 *								*
 *		-f#		copy only # files		*
 *								*
 ****************************************************************/


#include <stdio.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>
#include <sys/file.h>
#include <malloc.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "t-lib.h"

#define TRUE 1
#define FALSE 0

#define MAXBUF (63*1024)	/* maximum tape physical record size */
#define MOD512(var)  ((var) & (~511))	/* round down to next 512 byte bound */
int             maxbuf = MAXBUF;/* maximum tape physical record size (use) */

static char    *tape_name = "/dev/nrmt0";	/* default tape device */
static char    *aux_tape_name = "/dev/null";	/* default name for output tape */

static char     ffh[] = "t-io,f";	/* header for temp filename (data) */
static char     nfh[] = "t-io,n";	/* header for temp filename (control) */
static char     format[] = "%s.%d.%d";	/* format of filenames.
					 * prefix.tape.file */

#define FMTSIZ 200		/* limit on size of filenames created + 1 */
static int      tapen = 0;	/* current tape number */

static int      verbose = FALSE;/* true if we want messages */
static int      big_verbose = FALSE;	/* true if we want lots of messages */

enum {
    UN_INIT, M_READ, M_WRITE, M_COPY, M_ARCHIVE, M_UN_ARCHIVE
} 	 	inmode = UN_INIT;	/* mode */

int             convmode = 0;	/* conversational mode default to no */
char           *buffer;		/* buffer to use */

#define MAX_INT	0x7ffffff
static int      file_limit = MAX_INT;	/* copy as many files as there is tape */

#define PAD_SIZE 1024
#define PAD(buffer) ((buffer + PAD_SIZE -1) & (~(PAD_SIZE-1)))

static int      cartridge;	/* is this a cartridge tape */

/****************************************************************
 * usage -- bawl out the user					*
 ****************************************************************/
static void 
usage(void)
{
    fprintf(stderr, "Usage: t-io Control [options] Device1 [Device2]\n");
    fprintf(stderr, "Where:\n");
    fprintf(stderr, "	Control is one of\n");
    fprintf(stderr, "		-i	Copy device1 to disk\n");
    fprintf(stderr, "		-o	Copy disk to device1\n");
    fprintf(stderr, "		-c	Copy device1 to device2\n");
    fprintf(stderr, "		-A	Copy device1 to device2 archive fmt\n");
    fprintf(stderr, "		-U	Copy device1 to device2 un-archive\n");
    fprintf(stderr, " Options are\n");
    fprintf(stderr, "	-v	Verbose\n");
    fprintf(stderr, "	-V#	Very verbose, print message every # records\n");
    fprintf(stderr, "	-a	Ask user what to do about read errors\n");
    fprintf(stderr, "	-t#	Set tape number to # for disk file access\n");
    fprintf(stderr, "	-b#	Use # as blocking instead of 63k\n");
    fprintf(stderr, "	-f#	Copy only # files\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Device1 and Device2 may be in the form\n");
    fprintf(stderr, "		host:device\n");

    r_exit(2);
}
/****************************************************************
 * do_op -- perform a tape operation				*
 *								*
 * Paramters							*
 *	fd -- fd to use						*
 *	operation -- the operation to do			*
 *	count -- count of times to do it			*
 *								*
 * Returns							*
 *	0 -- ok							*
 *	-1 -- error						*
 ****************************************************************/
static int 
do_op(int fd, int operation, int count)
{
    struct mtop     mtop;	/* operation stuff if we use mtop */

    if (fd < 0)
	return (0);		/* ignore fake files */

    mtop.mt_count = count;
    mtop.mt_op = operation;

#ifdef DEBUG			/* allows copy to files (instead of tapes) */
    r_ioctl(fd, MTIOCTOP, &mtop);
    return (0);			/* always fake success */
#else	/* DEBUG */
    return (r_ioctl(fd, MTIOCTOP, &mtop));
#endif	/* DEBUG */
}
/****************************************************************
 * gen_name -- genearate a tape data file name			*
 *								*
 * Parameters							*
 *	file_number -- the number of the current file		*
 *								*
 * Returns							*
 *	the full name of the file				*
 ****************************************************************/
static char    *
gen_name(int file_number)
{
    static char     cfile[100];	/* must be static */

    sprintf(cfile, format, ffh, tapen, file_number);
    return (cfile);
}
/****************************************************************
 * read_check -- read from input fd and check for errors,	*
 *		will retry if the use wants us to		*
 *								*
 * Parameters							*
 *	fd -- fd of the file to read				*
 *	buffer -- the buffer to read into			*
 *	size -- the max size to read				*
 *	recno -- record number for error message		*
 *								*
 * Returns							*
 *	number of bytes read or -1 for error			*
 ****************************************************************/
static int 
read_check(int fd, char *read_buf, int size, long int recno)
{
    int             read_size;	/* size of read */
    static int      last_size = 0;	/* size of last read */
    char            buf[100];	/* buffer for answer */

    /* for debugging and consistancy */
    memset(read_buf, '\0', size);
    for (;;) {
	read_size = r_read(fd, read_buf, size);
	if (read_size >= 0) {
	    last_size = read_size;
	    return (read_size);
	}
	printf("Tape read error %d in record %ld\n", errno, recno);
	fflush(stdout);
	if (convmode) {
    ask_again:
	    printf("Abort, Retry, or Ignore? ");
	    if (fgets(buf, sizeof(buf), stdin) == NULL) {
		fprintf(stderr, "Unable to get answer\n");
		r_exit(8);
	    }
	    switch (*buf) {
	    case 'R':
	    case 'r':
		if (do_op(fd, MTBSR, 1) < 0) {
		    fprintf(stderr, "Can't backspace tape, aborting\n");
		    /* we don't know how to handle this */
		    r_exit(1);
		}
		continue;
	    case 'I':
	    case 'i':
		return (last_size);
	    case 'A':
	    case 'a':
		r_exit(1);
	    default:
		fprintf(stderr, "Bad response\n");
		goto ask_again;
	    }
	} else
	    return (-1);
    }
}
/****************************************************************
 * write_str -- printf a string into a fd			*
 *								*
 * Paramters							*
 *	fd -- fd to write to					*
 *	str -- string (printf style) for output			*
 *	parm1,2,3 -- parameters to printf			*
 ****************************************************************/
/*VARARGS2*/
void 
write_str(fd, str, parm1, parm2, parm3)
    int             fd;
    char           *str;
    int             parm1;
    int             parm2;
    int             parm3;
{
    char            out_str[1000];	/* the output string */
    int             out_len;	/* length of the string */

    if (fd < 0)
	return;			/* nothing to write to */

    sprintf(out_str, str, parm1, parm2, parm3);
    out_len = strlen(out_str);
    if (r_write(fd, out_str, out_len) != out_len) {
	perror("t-io: control write: ");
	r_exit(8);
    }
}
/****************************************************************
 * do_gen_read -- read the tape and do things to it		*
 *								*
 * Parameter							*
 *	input-fd -- fd of the input file			*
 *	data_fd -- fd to write data on (or -1 for do not write)	*
 *	control_fd -- fd to write control stuff on (-1 do not	*
 *			write)					*
 *	pad_flag -- true if we are to pad to next 1024 byte 	*
 *		boundry						*
 *	read_cartridge -- we are reading a cartridge		*
 *	file_number -- file number of the file to read		*
 *								*
 * Returns							*
 *	number of records in the data file read			*
 ****************************************************************/
static int 
do_gen_read(int input_fd, int data_fd, int control_fd,
	    int pad_flag, int read_cartridge, int file_number)
{
    int             read_size;	/* size of current read */
    int             last_size;	/* size of last read */
    int             recno;	/* number of current record */
    int             nrec;	/* number of records of current size */

    last_size = 0;
    /*
     * open the control disk file if needed
     */

    recno = nrec = 0;


    while (1) {
	read_size = read_check(input_fd, buffer, maxbuf, recno);

	if (read_size <= 0)
	    break;

	if (big_verbose != FALSE) {
	    if (recno % big_verbose == 0) {
		printf("Record %d size %d\n",
			      recno, read_size);
		fflush(stdout);
	    }
	}
	if (data_fd > 0) {
	    if (pad_flag)
		read_size = PAD(read_size);
	    if (r_write(data_fd, buffer, read_size) <= 0) {
		fprintf(stderr, "t-io: file write error #%d\n",
			       errno);
		perror("t-io: ");
		r_exit(2);
	    }
	}
	if (recno == 0) {
	    last_size = read_size;
	} else {

	    if (last_size != read_size) {
		write_str(control_fd, "%d,%d\n", recno - nrec, last_size);
		last_size = read_size;
		nrec = recno;
	    }
	}
	recno++;
    }

    if (read_size == 0) {
	write_str(control_fd, "%d,%d\n0,0\n", recno - nrec, last_size);
    }
    return (recno);
}

/****************************************************************
 * get_control_rec -- read data from control file and parse	*
 *								*
 * Parameters							*
 *	control_file -- file to read				*
 *	nrec -- number of records to read			*
 *	write_size -- size of the records			*
 ****************************************************************/
static void 
get_control_rec(FILE * control_file, int *nrec, int *write_size)
{
    char            buf[100];	/* buffer for data */

    if (fgets(buf, sizeof(buf), control_file) == NULL) {
	fprintf(stderr, "t-io: control file, malformed.\n");
	r_exit(2);
    }
    if (sscanf(buf, "%d,%d", nrec, write_size) != 2) {
	fprintf(stderr, "control file format error\n");
	r_exit(8);
    }
}

/****************************************************************
 * do_gen_write -- general purpose write routine		*
 *								*
 * Read data from the device and write it to the output		*
 * 	control file is honored					*
 *								*
 * Parameters							*
 *	input_fd -- the fd of the input file			*
 *	output_fd -- fd of the output file			*
 *	control_file -- file containing the control information	*
 *	pad_flag -- true if reads are to be padded		*
 ****************************************************************/
static int 
do_gen_write(int input_fd, int output_fd,
	     FILE * control_file, int pad_flag)
{
    int             write_size;	/* size of this write */
    int             recno;
    int             nrec;

    recno = 0;
    for (;;) {
	get_control_rec(control_file, &nrec, &write_size);

	if (nrec == 0) {
	    if (do_op(output_fd, MTWEOF, 1) != 0) {
		perror("t-io: eof write failed: ");
		r_exit(2);
	    }
	    return (recno);
	}
	/* write each record of this size */
	for (; nrec > 0; nrec--) {
	    int             act_write_size;

	    if (pad_flag)
		act_write_size = PAD(write_size);
	    else
		act_write_size = write_size;

	    if (r_read(input_fd, buffer, act_write_size) <= 0) {
		fprintf(stderr, "t-io: input file read error #%d\n",
			       errno);
		r_exit(2);
	    }
	    if (big_verbose != FALSE) {
		if (recno % big_verbose == 0) {
		    printf("Record %d size %d\n",
				  recno, write_size);
		    fflush(stdout);
		}
	    }
	    if (r_write(output_fd, buffer, write_size) <= 0) {
		fprintf(stderr, "t-io: tape write error #%d\n",
			       errno);
		r_exit(2);
	    }
	    recno++;
	}
    }
}
/****************************************************************
 * do_read -- process a read request				*
 *								*
 *  Read a file and write it to disk				*
 ****************************************************************/
static void 
do_read(void)
{
    int             ctl_fd;	/* fd of control file */
    int             data_fd = 0;/* fd of the data file */
    int             input_fd;	/* fd of the input file (tape) */
    char            cfile[FMTSIZ];	/* name of the control file */
    int             file_number;/* number of current file */

    input_fd = r_open(tape_name, 0);
    if (input_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", tape_name);
	r_exit(8);
    }

    sprintf(cfile, format, nfh, tapen, 0);

    ctl_fd = creat(cfile, 0666);
    if (ctl_fd < 0) {
	fprintf(stderr, "t-io: cannot open %s\n", cfile);
	r_exit(2);
    }
    for (file_number = 0; file_number < file_limit; file_number++) {
	data_fd = creat(gen_name(file_number), 0666);
	if (verbose) {
	    printf("Reading file %d\n", file_number);
	    fflush(stdout);
	}
	if (data_fd < 0) {
	    fprintf(stderr, "Error opening %s\n",
			   gen_name(file_number));
	    r_exit(8);
	}
	if (do_gen_read(input_fd, data_fd, ctl_fd,
			FALSE, cartridge, file_number) <= 0)
	    break;
	close(data_fd);
    }
    close(data_fd);
    unlink(gen_name(file_number));	/* unlink extra data file */
    close(ctl_fd);
    r_close(input_fd);
}
/****************************************************************
 * do_write -- read the disk and write the tape			*
 ****************************************************************/
static void 
do_write(void)
{
    FILE           *control_file;	/* control file */
    int             input_fd;	/* input file descriptor */
    int             output_fd;	/* output file descritor */
    int             file_number;/* number of current file */
    char            cfile[100];	/* file name of control file */

    /*
     * output mode - copy files back out to tape.
     */
    sprintf(cfile, format, nfh, tapen, 0);
    if ((control_file = fopen(cfile, "r")) == NULL) {
	fprintf(stderr, "t-io: cannot open %s\n", cfile);
	r_exit(2);
    }
    if ((output_fd = r_open(tape_name, 1)) < 0) {
	fprintf(stderr, "t-io: cannot open %s\n", tape_name);
	r_exit(2);
    }
    for (file_number = 0;; file_number++) {
	input_fd = open(gen_name(file_number), 0);
	if (input_fd < 0) {
	    int             nrec;
	    int             write_size;

	    get_control_rec(control_file, &nrec, &write_size);
	    if ((nrec == 0) && (write_size == 0))
		break;

	    fprintf(stderr, "Error opening %s\n",
			   gen_name(file_number));
	    r_exit(8);
	}
	if (verbose) {
	    printf("Writing file %d\n", file_number);
	    fflush(stdout);
	}
	if (do_gen_write(input_fd, output_fd, control_file, FALSE) == 0)
	    break;

	close(input_fd);
    }

    close(input_fd);
    fclose(control_file);
    if (do_op(output_fd, MTWEOF, 1) != 0) {
	perror("t-io: eof write failed: ");
	r_exit(2);
    }
    r_close(output_fd);
}
/****************************************************************
 * gen_copy -- generalized copy for data files			*
 *								*
 * Parameters							*
 *	input_fd -- the input file descriptor			*
 *	output_fd -- the output file descriptor			*
 *	control_fd -- the control file decrioptor		*
 *	pad -- true if we are to pad the output			*
 ****************************************************************/
static void 
gen_copy(int input_fd, int output_fd, int control_fd, int pad)
{
    int             file_number;/* number of current file */

    for (file_number = 0; file_number < file_limit; file_number++) {
	if (verbose) {
	    printf("Copying file %d\n", file_number);
	    fflush(stdout);
	}
	if (do_gen_read(input_fd, output_fd, control_fd, pad, cartridge, file_number) <= 0)
	    break;
	if (do_op(output_fd, MTWEOF, 1) != 0) {
	    perror("t-io: eof write failed: ");
	    r_exit(2);
	}
    }
    if (do_op(output_fd, MTWEOF, 1) != 0) {
	perror("t-io: eof write failed: ");
	r_exit(2);
    }
    if (do_op(output_fd, MTWEOF, 1) != 0) {
	perror("t-io: eof write failed: ");
	r_exit(2);
    }
    r_close(output_fd);
    r_close(input_fd);
}
/****************************************************************
 * do_copy -- copy from one tape to another			*
 ****************************************************************/
static void 
do_copy(void)
{
    int             output_fd;	/* fd of the output file */
    int             input_fd;	/* fd of the input file (tape) */

    input_fd = r_open(tape_name, 0);
    if (input_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", tape_name);
	r_exit(8);
    }

    output_fd = r_open(aux_tape_name, 1);
    if (output_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", aux_tape_name);
	r_exit(8);
    }
    gen_copy(input_fd, output_fd, -1, FALSE);

    r_close(output_fd);
    r_close(input_fd);
}
/****************************************************************
 * do_archive -- copy control and data to tape			*
 ****************************************************************/
static void 
do_archive(void)
{
    int             output_fd;	/* fd of the output file */
    int             input_fd;	/* fd of the input file (tape) */
    int             control_fd;	/* fd of control file */
    char           *tmp_name;	/* name of temp file */
    int             file_number;/* number of current file */

    input_fd = r_open(tape_name, 0);
    if (input_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", tape_name);
	r_exit(8);
    }

    output_fd = r_open(aux_tape_name, 1);
    if (output_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", aux_tape_name);
	r_exit(8);
    }
    tmp_name = tempnam(".", "t-io,tmp.");
    control_fd = open(tmp_name, O_RDWR | O_CREAT, 0666);	/* get control file */
    if (control_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", tmp_name);
	r_exit(8);
    }
    if (verbose) {
	printf("Scanning tape\n");
	fflush(stdout);
    }
    for (file_number = 0; file_number < file_limit; file_number++) {
	if (do_gen_read(input_fd, -1, control_fd, TRUE, cartridge, file_number) <= 0)
	    break;
    }


    if (lseek(control_fd, 0L, 0) < 0) {
	fprintf(stderr, " Control file rewind failed\n");
	r_exit(8);
    }
    /* copy control file */
    do_gen_read(control_fd, output_fd, -1, TRUE, FALSE, 0);

    close(control_fd);
    unlink(tmp_name);

    if (do_op(output_fd, MTWEOF, 1) != 0) {
	perror("t-io: eof write failed: ");
	r_exit(2);
    }
    if (verbose) {
	printf("Copying data\n");
	fflush(stdout);
    }
    if (do_op(input_fd, MTREW, 1) != 0) {
	perror("t-io: rewind failed: ");
	r_exit(2);
    }
    gen_copy(input_fd, output_fd, -1, TRUE);
}


/****************************************************************
 * do_unarchive -- undo the work of archive			*
 ****************************************************************/
static void 
do_unarchive(void)
{
    int             output_fd;	/* fd of the output file */
    int             input_fd;	/* fd of the input file  */
    int             control_fd;	/* fd of the control file */
    char           *tmp_name;	/* name of the tmp file */
    int             file_number;/* name of current file */
    FILE           *control_file;	/* control file */

    input_fd = r_open(tape_name, 0);
    if (input_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", tape_name);
	r_exit(8);
    }
    output_fd = r_open(aux_tape_name, 1);
    if (output_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", aux_tape_name);
	r_exit(8);
    }
    tmp_name = tempnam(".", "t-io,tmp.");
    control_fd = creat(tmp_name, 0666);	/* get control file */
    if (control_fd < 0) {
	fprintf(stderr, "Unable to open %s\n", tmp_name);
	r_exit(8);
    }
    if (verbose) {
	printf("Copying control file\n");
	fflush(stdout);
    }
    if (do_gen_read(input_fd, control_fd, -1, FALSE, cartridge, 0) <= 0) {
	fprintf(stderr, "Problem copying control file\n");
	r_exit(8);
    }
    close(control_fd);
    control_file = fopen(tmp_name, "r");
    if (control_file == NULL) {
	fprintf(stderr, "Unable to open %s\n", tmp_name);
	r_exit(8);
    }
    if (verbose) {
	printf("Copying data\n");
	fflush(stdout);
    }
    for (file_number = 0;; file_number++) {
	if (verbose) {
	    printf("Writing file %d\n", file_number);
	    fflush(stdout);
	}
	if (do_gen_write(input_fd, output_fd, control_file, TRUE) == 0)
	    break;
	if (do_op(input_fd, MTFSF, 1) != 0) {
	    perror("t-io: fsf failed: ");
	    r_exit(2);
	}
    }

    r_close(output_fd);
    r_close(input_fd);
    fclose(control_file);
    unlink(tmp_name);
}
int 
main(int argc, char *argv[])
{
    char           *p;		/* current option character */

    argv++;			/* move past the program name */

    while ((argc > 1) && (argv[0][0] == '-')) {
	p = *argv;
	p++;

	while (*p != '\0') {
	    switch (*p++) {
		/* i -- copy tape to disk */
	    case 'i':
		if (inmode != UN_INIT)
		    usage();;
		inmode = M_READ;
		break;
		/* o -- copy disk to tape */
	    case 'o':
		if (inmode != UN_INIT)
		    usage();
		inmode = M_WRITE;
		break;
		/* c -- copy tape to tape */
	    case 'c':
		if (inmode != UN_INIT)
		    usage();
		inmode = M_COPY;
		break;
		/* A -- copy tape to tape (archive) */
	    case 'A':
		if (inmode != UN_INIT)
		    usage();
		inmode = M_ARCHIVE;
		break;
		/* U -- copy tape to tape (un-archive) */
	    case 'U':
		if (inmode != UN_INIT)
		    usage();
		inmode = M_UN_ARCHIVE;
		break;

		/* ----------------------------------------- */
		/* b# specify blocking */
	    case 'b':
		maxbuf = atoi(&p[0]);
		(*p) = '\0';
		break;
	    case 'a':
		convmode = TRUE;
		break;
		/* v -- verbose */
	    case 'v':
		verbose = TRUE;
		break;
	    case 'V':
		big_verbose = atoi(&p[0]);
		if (big_verbose == 0)
		    big_verbose = 1;
		(*p) = 0;
		break;
	    case 't':
		tapen = atoi(&p[0]);
		(*p) = 0;
		break;
	    case 'f':
		file_limit = atoi(&p[0]);
		(*p) = '\0';
		break;
	    default:
		usage();
	    }
	}
	argc--;
	argv++;
    }

    if (argc > 1) {
	tape_name = argv[0];
	argv++;
	argc--;
    }
    if (argc > 1) {
	aux_tape_name = argv[0];
	argv++;
	argc--;
    }
    if (argc > 1)
	usage();

    buffer = malloc((unsigned) (PAD(maxbuf)));	/* allocate buffer */


    switch (inmode) {
    case M_READ:
	do_read();
	break;
    case M_WRITE:
	do_write();
	break;
    case M_COPY:
	do_copy();
	break;
    case M_ARCHIVE:
	do_archive();
	break;
    case M_UN_ARCHIVE:
	do_unarchive();
	break;
    default:
	usage();
    }
    r_exit(0);
    return (0);
}

